
#include <cstdio>
#include <cstdlib>
#include <cmath>

static const int ayLevelTable[16] = {
      0,   300,   447,   635,   925,  1351,  1851,  2991,
   3695,  5782,  7705,  9829, 12460, 15014, 18528, 21845
};

static void fft(float *buf, size_t n, bool isInverse)
{
  // check FFT size
  if (n < 16 || n > 65536 || (n & (n - 1)) != 0)
    return;
  if (!isInverse) {
    // convert real data to interleaved real/imaginary format
    size_t  i = n;
    do {
      i--;
      buf[(i << 1) + 0] = buf[i];
      buf[(i << 1) + 1] = 0.0f;
    } while (i);
  }
  else {
    buf[1] = 0.0f;
    buf[n + 1] = 0.0f;
    for (size_t i = 1; i < (n >> 1); i++) {
      buf[((n - i) << 1) + 0] = buf[(i << 1) + 0];
      buf[((n - i) << 1) + 1] = -(buf[(i << 1) + 1]);
    }
  }
  // pack data in reverse bit order
  size_t  i, j;
  for (i = 0, j = 0; i < n; i++) {
    if (i < j) {
      float   tmp1 = buf[(i << 1) + 0];
      float   tmp2 = buf[(i << 1) + 1];
      buf[(i << 1) + 0] = buf[(j << 1) + 0];
      buf[(i << 1) + 1] = buf[(j << 1) + 1];
      buf[(j << 1) + 0] = tmp1;
      buf[(j << 1) + 1] = tmp2;
    }
    for (size_t k = (n >> 1); k > 0; k >>= 1) {
      j ^= k;
      if ((j & k) != 0)
        break;
    }
  }
  // calculate FFT
  for (size_t k = 1; k < n; k <<= 1) {
    double  ph_inc_re = std::cos(4.0 * std::atan(1.0) / double(long(k)));
    double  ph_inc_im = std::sin((isInverse ? 4.0 : -4.0)
                                 * std::atan(1.0) / double(long(k)));
    for (j = 0; j < n; j += (k << 1)) {
      double  ph_re = 1.0;
      double  ph_im = 0.0;
      for (i = j; i < (j + k); i++) {
        double  re1 = buf[(i << 1) + 0];
        double  im1 = buf[(i << 1) + 1];
        double  re2 = buf[((i + k) << 1) + 0];
        double  im2 = buf[((i + k) << 1) + 1];
        buf[(i << 1) + 0] = float(re1 + re2 * ph_re - im2 * ph_im);
        buf[(i << 1) + 1] = float(im1 + re2 * ph_im + im2 * ph_re);
        buf[((i + k) << 1) + 0] = float(re1 - re2 * ph_re + im2 * ph_im);
        buf[((i + k) << 1) + 1] = float(im1 - re2 * ph_im - im2 * ph_re);
        double  tmp = ph_re * ph_inc_re - ph_im * ph_inc_im;
        ph_im = ph_re * ph_inc_im + ph_im * ph_inc_re;
        ph_re = tmp;
      }
    }
  }
  if (!isInverse) {
    buf[1] = 0.0f;
    buf[n + 1] = 0.0f;
  }
  else {
    // convert from interleaved real/imaginary format to pure real data
    for (i = 0; i < n; i++)
      buf[i] = buf[(i << 1) + 0];
    for (i = n; i < (n << 1); i++)
      buf[i] = 0.0f;
  }
}

static double calculateDistortion(const unsigned char *dacTable)
{
  float   fftBuf[16384];
  std::FILE *f = std::fopen("dac.raw", "wb");
  for (int i = 0; i < 16384; i++)
    fftBuf[i] = 0.0f;
  int     minVal = 32767;
  int     maxVal = -32768;
  int     a = 0;
  int     b = 0;
  int     c = 0;
  int     y = 0;
  for (int i = 0; i < 16384; i++) {
    switch (i & 7) {
    case 0:
      y = int(std::sin(double(i >> 3) * (8.0 * std::atan(1.0) / 1024.0)) * 95.5
              + 96.0);
      a = ayLevelTable[dacTable[y]];
      break;
    case 2:
      b = ayLevelTable[dacTable[y + 257]];
      break;
    case 4:
      c = ayLevelTable[dacTable[y + 512]];
      break;
    }
    int     tmp = a + b + c;
    if (i < 8192) {
      if (i >= 8) {
        minVal = (tmp < minVal ? tmp : minVal);
        maxVal = (tmp > maxVal ? tmp : maxVal);
      }
    }
    else {
      tmp = int((double(tmp) + (double(minVal - maxVal) * 0.5))
                * (65534.0 / double(maxVal - minVal)) + 32768.5) - 32768;
      fftBuf[i & 8191] = float(tmp);
      if (f) {
        std::fputc(tmp & 0xFF, f);
        std::fputc((tmp & 0xFF00) >> 8, f);
      }
    }
  }
  if (f)
    std::fclose(f);
  fft(&(fftBuf[0]), 8192, false);
  double  r = 0.0;
  double  d = 0.0;
  for (int i = 1; i <= 4096; i++) {
    float   re = fftBuf[i << 1] * (1.0f / 8192.0f);
    float   im = fftBuf[(i << 1) + 1] * (1.0f / 8192.0f);
    if (i == 1) {
      r = double(re * re) + double(im * im);
    }
    else {
      d = d + ((double(re * re) + double(im * im))
               * (double(i < 10 ? i : 10) * 0.5));
    }
  }
  d = std::sqrt(d / r) * 100.0;
  return d;
}

int main(int argc, char **argv)
{
  unsigned char dacTable[768];
  for (int i = 0; i < 768; i++)
    dacTable[i] = 0;
  int     maxC = 6;
  if (argc > 2)
    maxC = std::atoi(argv[2]);
  std::FILE *f = std::fopen("dactable.txt", "w");
  if (maxC >= 1 && maxC <= 15) {
    int     lvlMult = 239;              // for maxC == 6
    if (argc > 1)
      lvlMult = std::atoi(argv[1]);
    int     bestA = 0;
    int     bestB = 0;
    int     bestC = 0;
    for (int i = 0; i < 192; i++) {
      int     v = i * lvlMult;
      double  minErr = 1.0e24;
      for (int a = bestA; a < 16; a++) {
        for (int b = bestB; b < 16; b++) {
          for (int c = 0; c <= maxC; c++) {
            int     v_ = ayLevelTable[a] + ayLevelTable[b] + ayLevelTable[c];
            double  err = double(v_ - v) / double(191 * lvlMult);
            err = err * err;
            if (err < minErr) {
              minErr = err;
              bestA = a;
              bestB = b;
              bestC = c;
            }
          }
        }
      }
      dacTable[i] = (unsigned char) bestA;
      dacTable[i + 257] = (unsigned char) bestB;
      dacTable[i + 512] = (unsigned char) bestC;
      if (f) {
        std::fprintf(f,
                     "%3d: %2d %2d %2d (%5d %5d %5d), %6.4f (%6.4f), "
                     "err = %6.4f\n",
                     i, bestA, bestB, bestC, ayLevelTable[bestA],
                     ayLevelTable[bestB], ayLevelTable[bestC],
                     double(i) / 191.0,
                     double(ayLevelTable[bestA] + ayLevelTable[bestB]
                            + ayLevelTable[bestC]) / double(191 * lvlMult),
                     std::sqrt(minErr));
      }
    }
  }
  else {
    int     lvlMult = 250;
    if (argc > 1)
      lvlMult = std::atoi(argv[1]);
    int     prvB = 0;
    int     prvC = 0;
    for (int i = 0; i < 192; i++) {
      int     v = i * lvlMult;
      int     bestA = 0;
      int     bestB = 0;
      int     bestC = 0;
      double  minErr = 1.0e24;
      for (int a = 0; a < 16; a++) {
        for (int b = 0; b < 16; b++) {
          for (int c = 0; c < 16; c++) {
            int     v_, v1, v2, v3;
            v_ = (i - int(bool(i))) * lvlMult;
            v1 = ayLevelTable[a] + ayLevelTable[prvB] + ayLevelTable[prvC];
            v2 = ayLevelTable[a] + ayLevelTable[b] + ayLevelTable[prvC];
            v3 = ayLevelTable[a] + ayLevelTable[b] + ayLevelTable[c];
            double  err = double(v3 - v);
            double  err1 = double(v1) - (double(v + v_ + v_) * (1.0 / 3.0));
            double  err2 = double(v2) - (double(v + v + v_) * (1.0 / 3.0));
            err = err * (1.0 / double(191 * lvlMult));
            err1 = err1 * (1.0 / double(191 * lvlMult));
            err2 = err2 * (1.0 / double(191 * lvlMult));
            err = (err * err * 0.8) + (err1 * err1 * 0.1) + (err2 * err2 * 0.1);
            if (err < minErr) {
              minErr = err;
              bestA = a;
              bestB = b;
              bestC = c;
            }
          }
        }
      }
      dacTable[i] = (unsigned char) bestA;
      dacTable[i + 257] = (unsigned char) bestB;
      dacTable[i + 512] = (unsigned char) bestC;
      prvB = bestB;
      prvC = bestC;
      if (f) {
        std::fprintf(f,
                     "%3d: %2d %2d %2d (%5d %5d %5d), %6.4f (%6.4f), "
                     "err = %6.4f\n",
                     i, bestA, bestB, bestC, ayLevelTable[bestA],
                     ayLevelTable[bestB], ayLevelTable[bestC],
                     double(i) / 191.0,
                     double(ayLevelTable[bestA] + ayLevelTable[bestB]
                            + ayLevelTable[bestC]) / double(191 * lvlMult),
                     std::sqrt(minErr));
      }
    }
  }
  if (f)
    std::fclose(f);
  f = std::fopen("dactable.s", "w");
  if (f) {
    std::fprintf(f, "\n        align 256\n\ndacTable:\n");
    for (int i = 0; i < 768; i++) {
      if (!(i & 15))
        std::fprintf(f, "        defb  ");
      std::fprintf(f, "%2d", int(dacTable[i]));
      if ((i & 15) != 15)
        std::fprintf(f, ", ");
      else
        std::fprintf(f, "\n");
    }
    std::fprintf(f, "\n");
    std::fclose(f);
  }
  std::printf("Distortion = %.3f%%\n", calculateDistortion(&(dacTable[0])));
  return 0;
}

